/**HEADER********************************************************************
* 
* Copyright (c) 2008 Freescale Semiconductor;
* All Rights Reserved
*
*************************************************************************** 
*
* THIS SOFTWARE IS PROVIDED BY FREESCALE "AS IS" AND ANY EXPRESSED OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
* IN NO EVENT SHALL FREESCALE OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
* THE POSSIBILITY OF SUCH DAMAGE.
*
**************************************************************************
*
* $FileName: adc.c$
* $Version : 3.0.12.0$
* $Date    : Jun-29-2009$
*
* Comments:
*
*   This file contains the generic ADC driver functions
*
*END************************************************************************/


#include "mqx.h"
#include "fio.h"
#include "fio_prv.h"
#include "io.h"
#include "adc.h"
#include "adc_prv.h"
#include "adc_conf.h"

#if BSPCFG_ENABLE_ADC

/* configuration of driver, channel(s) independent */
ADC_PTR adc = NULL;
/* channel dependent configurations */
ADC_CHANNEL_PTR adc_ch[ADC_MAX_CHANNELS] = { 0 };

/* scheduler constants */
static ADC_TRIGGER_MASK running_mask = 0;

/*FUNCTION*****************************************************************
* 
* Function Name    : _adc_install
* Returned Value   : _mqx_uint a task error code or MQX_OK
* Comments         :
*    Install a adc driver.
*
*END*********************************************************************/

_mqx_uint _io_adc_install
   (
      /* [IN] A string that identifies the device for fopen */
      char_ptr  identifier
   ) 
   
{ /* Body */
   _mqx_uint result;
   
   result = _io_dev_install(identifier,
      _adc_open,
      _adc_close,
      _adc_read,
      _adc_write,
      _adc_ioctl,
      NULL); 

      return result;

} /* Endbody */


/*FUNCTION*****************************************************************
* 
* Function Name    : _adc_open
* Returned Value   : _mqx_int - IO_ERROR or IO_OK
* Comments         : Opens and initializes ADC driver.
* 
*END*********************************************************************/

_mqx_int _adc_open
   (
      /* [IN] the file handle for the device being opened */
      FILE_PTR   fd_ptr,
       
      /* [IN] the remaining portion of the name of the device */
      char_ptr   open_name_ptr,

      /* [IN] the initialization parameters (ADC_INIT_STRUCT_PTR) */
      char_ptr   flags
   )
{ /* Body */
   char_ptr file_name_ptr = fd_ptr->DEV_PTR->IDENTIFIER; /* use driver name temporarily */
   _mqx_int status;

   while (*file_name_ptr++ != 0)
      open_name_ptr++; /* move to the file name */
   file_name_ptr = open_name_ptr; /* throw out file_name_ptr and assign the correct file name */

   if (*file_name_ptr == 0) {
      /* 
      ** ADC DEVICE DRIVER OPENED WITHOUT CHANNEL SPECIFIED
      */
      ADC_INIT_STRUCT_PTR init_from = (ADC_INIT_STRUCT_PTR) flags;

      if (init_from == NULL)
         return IO_ERROR; /* mandatory open param missing */

      if (NULL == adc) {
         adc = (ADC_PTR) _mem_alloc_zero(sizeof(ADC));
#if MQX_CHECK_MEMORY_ALLOCATION_ERRORS
         if (adc == NULL)
            return IO_ERROR; /* memory alllocation error */
#endif
      } else
         return IO_ERROR; /* cannot open more than one instance of ADC driver */
      
      adc->g.resolution = init_from->resolution;
      adc->g.triggers_fire = 0; /* no trigger fired */
      adc->g.run = 0;

      if (IO_OK != (status = _adc_hw_init(adc))) { /* init cpu according to adc_general_config */
         _mem_free(adc);
         adc = NULL;
      }
      
      //fd_ptr->DEV_DATA_PTR = NULL; /* not needed: we dont need to store any additional data */

      return status;
   }
   else {
      /*
      ** CHANNEL TO BE OPENED, NUMBER IS IN OPEN_NAME_PTR
      */
      ADC_INIT_CHANNEL_STRUCT_PTR init_from = (ADC_INIT_CHANNEL_STRUCT_PTR) flags;
      _mqx_uint                   ch = 0;
      _mqx_uint                   radix = 1;

      if (init_from == NULL)
         return IO_ERROR; /* mandatory open param missing */

      if (NULL == adc)
         return IO_ERROR; /* general device was not opened */

      if (init_from->number_samples < 1)
         return IO_ERROR; /* number of samples must be more than 0 */

      /* move through digits to the end of string */
//      open_name_ptr = file_name_ptr; // not needed, already done
      while ((*open_name_ptr >= '0') && (*open_name_ptr <= '9'))
         open_name_ptr++;

      if (*open_name_ptr != 0) /* if any non-digit character was found */
         return IO_ERROR;

      open_name_ptr--; /* move to the last digit in string */

      do {
         if (ADC_MAX_CHANNELS <= (ch += radix * (*open_name_ptr - '0')))
            return IO_ERROR; /* the number of channel is too high */
         radix *= 10;
      } while (open_name_ptr-- != file_name_ptr); /* move in the string to the start of channel number */

      /* channel_number now contains valid channel number */

      if (adc_ch[ch] == NULL) {
         adc_ch[ch] = (ADC_CHANNEL_PTR) _mem_alloc_zero(sizeof(ADC_CHANNEL));
#if MQX_CHECK_MEMORY_ALLOCATION_ERRORS
         if (adc_ch[ch] == NULL)
            return IO_ERROR; /* memory allocation error */
#endif
      } else
         return IO_ERROR; /* channel already in use */

      /* allocate channel buffer */
      adc_ch[ch]->g.buffer_start = _mem_alloc_zero((init_from->buffer_size + 1) * sizeof(ADC_RESULT_STRUCT)); /* zero is needed for ADC_CHANNEL_ACCUMULATE */
      adc_ch[ch]->g.buffer_end = adc_ch[ch]->g.buffer_start + (init_from->buffer_size + 1);

      if (adc_ch[ch]->g.buffer_start == NULL) {
         _mem_free (adc_ch[ch]);
         adc_ch[ch] = NULL;
         return IO_ERROR; /* memory allocation error */
      }

      // initialize channel
      adc_ch[ch]->g.buffer_task = adc_ch[ch]->g.buffer_start;
      adc_ch[ch]->g.buffer_driver = adc_ch[ch]->g.buffer_start;

      adc_ch[ch]->g.number = ch;
      adc_ch[ch]->g.source = init_from->source;
      adc_ch[ch]->g.init_flags = init_from->flags;
      adc_ch[ch]->g.trigger = init_from->trigger;
      adc_ch[ch]->g.offset = init_from->time_offset;
      adc_ch[ch]->g.period = init_from->time_period;
      
      
      adc_ch[ch]->g.samples = 0;
      adc_ch[ch]->g.samples_preset = init_from->number_samples - 1;
#if MQX_USE_LWEVENTS
      adc_ch[ch]->g.complete_event = init_from->complete_event;
#endif

      if (IO_OK != (status = _adc_hw_channel_init(ch))) { /* this should initialize ADC hardware */
           _mem_free(adc_ch[ch]->g.buffer_start);
          _mem_free (adc_ch[ch]);
          adc_ch[ch] = NULL;
          return status;
      }

      /* store file context for opened channel */
      fd_ptr->DEV_DATA_PTR = adc_ch[ch]; /* store information about channel */

      /* if channel should be run in opening file phase, then run it using special trigger */
      if (!(init_from->flags & ADC_CHANNEL_START_TRIGGERED)) {
          return _adc_ioctl(fd_ptr, IOCTL_ADC_RUN_CHANNEL, NULL);
      }
   }

   return IO_OK;

} /* Endbody */


/*FUNCTION*****************************************************************
* 
* Function Name    : _adc_close
* Returned Value   : ERROR CODE
* Comments         : Closes ADC driver or channel
* 
*END*********************************************************************/

_mqx_int _adc_close
   (
      /* [IN] the file handle for the device being closed */
      FILE_PTR   fd_ptr
   )
{ /* Body */
   ADC_CHANNEL_GENERIC_PTR channel = (ADC_CHANNEL_GENERIC_PTR) fd_ptr->DEV_DATA_PTR;
   _mqx_int        i, status;

   for (i = 0, status = 0; i < ADC_MAX_CHANNELS; i++) {
      if (adc_ch[i] != NULL) {
         status++;
         if ((pointer) adc_ch[i] == (pointer) channel) {
            _adc_stop(channel, 0);
            if (IO_OK != (status = _adc_hw_channel_deinit(channel)))
               return status;
            _mem_free(channel->buffer_start);
            _mem_free(channel);
            adc_ch[i] = NULL;
            return IO_OK;
         }
      }
   }
   
   /* if we reach this code, then we close ADC */
   /* test, if there is still any open channel */
   if (status != 0)
         return IO_ERROR; /* cannot close ADC driver: there are still open channels */

   if (IO_OK != (status = _adc_hw_deinit()))
      return status;
   _mem_free(adc);
   adc = NULL;

   return IO_OK;

} /* Endbody */


/*FUNCTION*****************************************************************
* 
* Function Name    : _adc_read
* Returned Value   : number of characters read
* Comments         : Reads data from ADC driver
* 
*END*********************************************************************/

_mqx_int _adc_read
   (
      /* [IN] the file handle for the device */
      FILE_PTR   fd_ptr,

      /* [IN] where the characters are to be stored */
      char_ptr   data_ptr,

      /* [IN] the number of characters to input */
      _mqx_int   num
   )
{ /* Body */
    _mqx_int cp_cnt;
    uint_32_ptr dst = (uint_32_ptr)data_ptr;
    ADC_CHANNEL_GENERIC_PTR ch_conf = (ADC_CHANNEL_GENERIC_PTR) fd_ptr->DEV_DATA_PTR;
    
    if (ch_conf->buffer_task == NULL)
        return 0; /* channel not run yet */
        
    cp_cnt = (ch_conf->buffer_driver - ch_conf->buffer_task);
    if (cp_cnt < 0)
        cp_cnt += (ch_conf->buffer_end - ch_conf->buffer_start);
    
    if (cp_cnt > (num / sizeof(uint_32)))
        cp_cnt = (num / sizeof(uint_32));
    
    while (cp_cnt) {
        *dst++ = ch_conf->buffer_task->result;
        ch_conf->buffer_task->result = 0; /* zero buffer to be usable next time (applicable for ADC_CHANNEL_ACCUMULATE) */
        if (++ch_conf->buffer_task >= ch_conf->buffer_end)
            ch_conf->buffer_task = ch_conf->buffer_start;
        
        cp_cnt--;
    }
    
    return (char_ptr)dst - data_ptr;
} /* Endbody */


/*FUNCTION*****************************************************************
* 
* Function Name    : _adc_write
* Returned Value   : number of characters written
* Comments         : Does nothing, ADC is solely read-only driver
* 
*END*********************************************************************/

_mqx_int _adc_write
   (
      /* [IN] the file handle for the device */
      FILE_PTR   fd_ptr,

      /* [IN] where the characters are */
      char_ptr   data_ptr,

      /* [IN] the number of characters to output */
      _mqx_int   num
   )
{ /* Body */
   return 0;
} /* Endbody */


/*FUNCTION*****************************************************************
* 
* Function Name    : _adc_ioctl
* Returned Value   : int_32
* Comments         :
*    Returns result of ioctl operation.
*
*END*********************************************************************/

_mqx_int _adc_ioctl
   (
      /* [IN] the file handle for the device */
      FILE_PTR   fd_ptr,

      /* [IN] the ioctl command */
      _mqx_uint  cmd,

      /* [IN] the ioctl parameters */
      pointer    param_ptr
   )
{ /* Body */
	ADC_CHANNEL_GENERIC_PTR adc_ch = fd_ptr->DEV_DATA_PTR;
	uint_32_ptr             uparam_ptr = param_ptr;

	switch (cmd) {
		case IOCTL_ADC_RUN_CHANNEL:
			/* Runs specific channel */
			return _adc_trigger(adc_ch, 0);
		case IOCTL_ADC_FIRE_TRIGGER:
			/* Runs channels associated with trigger mask */
			return _adc_trigger(NULL, (ADC_TRIGGER_MASK)param_ptr);
		case IOCTL_ADC_STOP_CHANNEL:
			/* Stops specific channel */
			return _adc_stop(adc_ch, 0);
		case IOCTL_ADC_STOP_CHANNELS:
			/* Stops channels associated with trigger mask */
			return _adc_stop(NULL, (ADC_TRIGGER_MASK)param_ptr);
		case IOCTL_ADC_PAUSE_CHANNEL:
			/* Stops specific channel */
			return _adc_pause(adc_ch, 0);
		case IOCTL_ADC_PAUSE_CHANNELS:
			/* Stops channels associated with trigger mask */
			return _adc_pause(NULL, (ADC_TRIGGER_MASK)param_ptr);
		case IOCTL_ADC_RESUME_CHANNEL:
			/* Resumes specific channel */
			return _adc_resume(adc_ch, 0);
		case IOCTL_ADC_RESUME_CHANNELS:
			/* Resumes channels associated with trigger mask */
			return _adc_resume(NULL, (ADC_TRIGGER_MASK)param_ptr);
		case IO_IOCTL_DEVICE_IDENTIFY:	
			/* return the device identify */
			uparam_ptr[0] = IO_DEV_TYPE_PHYS_ADC;
			uparam_ptr[1] = 0;
			uparam_ptr[2] = IO_DEV_ATTR_READ;
			return MQX_OK;			
	}
    return IO_ERROR_INVALID_IOCTL_CMD;
} /* Endbody */

/*FUNCTION*****************************************************************
* 
* Function Name    : _adc_scheduler
* Returned Value   : 
* Comments         : Schedules next ADC conversion and informs low-level
*                    driver about that time. Low-level driver must call
*                    this scheduler in that time.
*                    This function runs in interrupt!
*END*********************************************************************/

void _adc_scheduler(_mqx_uint time) { /* gets actual time as parameter */
    _mqx_int i;
    _mqx_uint wait;
    _mqx_int rem;
    
    if (running_mask) { /* if there is any channel in the mask for measurement */
        _adc_hw_measure(running_mask); /* measure channels */
    
        for (i = 0; i < ADC_MAX_CHANNELS; i++) { /* update time of last measurement in channels */
            if (!adc_ch[i] || !(running_mask & (1 << adc_ch[i]->g.source)))
                continue;
            
            adc_ch[i]->g.time = time;
        }
    }
    
    /* look for closest AD conversion */
    for (i = 0, wait = -1, running_mask = 0; i < ADC_MAX_CHANNELS; i++) {
        if (adc_ch[i] && adc_ch[i]->g.runtime_flags & ADC_CHANNEL_RUNNING) {
            /* compute remaining time to next measurement:
            ** elapsed time from last measurement: time - adc_ch[i]->time
            ** remaining time to next measurement: adc_ch[i]->period - elapsed
            */
            rem = adc_ch[i]->g.period - (time - adc_ch[i]->g.time);
        
            if (rem < 0)
                rem = 0;
        
            if (rem < wait) {
                wait = rem;
                running_mask = (1 << adc_ch[i]->g.source);
            }
            else if (rem == wait) {
                running_mask |= (1 << adc_ch[i]->g.source);
            }
        }
    }
    
    /* test if any conversion should be performed next time */
    if (running_mask) {
        if (!adc->g.run) {
            _adc_hw_run();
            adc->g.run = 1;
        }
    }
    else {
        _adc_hw_stop();
        adc->g.run = 0;
    }
    
    /* inform low level driver that next dispatch call should be done in 'wait' time */
    _adc_hw_set_wait(wait);
}

/*FUNCTION*****************************************************************
* 
* Function Name    : _adc_conv_end
* Returned Value   : 
* Comments         : This is called by low-level driver to inform generic
*                    driver, that some conversion was performed. Function
*                    moves stored data to the channel buffers.
*END*********************************************************************/

void _adc_conv_end(uint_32 *results_buf, ADC_HW_CHANNEL_MASK mask) {
    _mqx_int i;

    for (i = 0; i < ADC_MAX_CHANNELS; i++) {
        if (!adc_ch[i] || !(adc_ch[i]->g.runtime_flags & ADC_CHANNEL_RESUMED) || !(mask & (1 << adc_ch[i]->g.source)))
            continue;
            
        if (adc_ch[i]->g.init_flags & ADC_CHANNEL_ACCUMULATE)
            /* if the channel has ADC_CHANNEL_ACCUMULATE, then accumulate result */
            adc_ch[i]->g.buffer_driver->result += results_buf[adc_ch[i]->g.source];

        else
            /* else copy the result to the channel buffer */
            adc_ch[i]->g.buffer_driver->result = results_buf[adc_ch[i]->g.source];

        if (!(adc_ch[i]->g.init_flags & ADC_CHANNEL_ACCUMULATE) || !(adc_ch[i]->g.samples)) {
            /* move driver pointer to next index, circullarily */
            if (++adc_ch[i]->g.buffer_driver >= adc_ch[i]->g.buffer_end)
                adc_ch[i]->g.buffer_driver = adc_ch[i]->g.buffer_start;

            /* if new value should overwrite old value in ring buffer, move read pointer to next value */
            if (adc_ch[i]->g.buffer_driver == adc_ch[i]->g.buffer_task) {
                if (++adc_ch[i]->g.buffer_task >= adc_ch[i]->g.buffer_end)
                    adc_ch[i]->g.buffer_task = adc_ch[i]->g.buffer_start;
            }
        }
        
        if (adc_ch[i]->g.samples) {
            adc_ch[i]->g.samples--;
        }
        else {
#if MQX_USE_LWEVENTS
            /* if it is the last sample in the batch, then set event */
            if (adc_ch[i]->g.complete_event)
                _lwevent_set(adc_ch[i]->g.complete_event, 1); /* set completion event */
#endif
                
            if (adc_ch[i]->g.init_flags & ADC_CHANNEL_MEASURE_ONCE)
                adc_ch[i]->g.runtime_flags &= ~ADC_CHANNEL_RUNNING;
            
            adc_ch[i]->g.samples = adc_ch[i]->g.samples_preset;
        }
    }
}

/*FUNCTION*****************************************************************
* 
* Function Name    : _adc_trigger
* Returned Value   : IO_OK
* Comments         : Starts channel measurement or more channels defined
*                    by trigger mask
*
*END*********************************************************************/

_mqx_int _adc_trigger(ADC_CHANNEL_GENERIC_PTR channel, ADC_TRIGGER_MASK mask) {
    _mqx_int i;
    _mqx_uint time;

	_int_disable();
	time = _adc_hw_get_time();
    
    if (channel) {
        channel->time = time + channel->offset;
        channel->samples = channel->samples_preset;
	    channel->runtime_flags |= ADC_CHANNEL_RUNNING | ADC_CHANNEL_RESUMED;
    }
    else {
        for (i = 0; i < ADC_MAX_CHANNELS; i++) {
            if (adc_ch[i] && (adc_ch[i]->g.trigger & mask)) {
                adc_ch[i]->g.time = time + adc_ch[i]->g.offset;
            	/* period was subtracted because next scheduling will add it */
                adc_ch[i]->g.samples = adc_ch[i]->g.samples_preset;
	            adc_ch[i]->g.runtime_flags |= ADC_CHANNEL_RUNNING | ADC_CHANNEL_RESUMED;
            }
        }
    }
    
    running_mask = 0;
    _adc_scheduler(time);

    _int_enable();
    return IO_OK;
}

/*FUNCTION*****************************************************************
* 
* Function Name    : _adc_pause
* Returned Value   : IO_OK
* Comments         : Pauses measurement of single channel, or more channels
*                    defined by trigger mask
*
*END*********************************************************************/

_mqx_int _adc_pause(ADC_CHANNEL_GENERIC_PTR channel, ADC_TRIGGER_MASK mask) {
    _mqx_int i;
    _mqx_uint time;

	_int_disable();

    if (channel) {
	    channel->runtime_flags &= ~ADC_CHANNEL_RESUMED;
    }
    else {
        for (i = 0; i < ADC_MAX_CHANNELS; i++) {
            if (adc_ch[i] && (adc_ch[i]->g.trigger & mask)) {
	            adc_ch[i]->g.runtime_flags &= ~ADC_CHANNEL_RESUMED;
            }
        }
    }

    _int_enable();
    return IO_OK;
}

/*FUNCTION*****************************************************************
* 
* Function Name    : _adc_trigger
* Returned Value   : IO_OK
* Comments         : Resumes measurement of single paused channel, or more
*                    channels defined by trigger mask
*
*END*********************************************************************/

_mqx_int _adc_resume(ADC_CHANNEL_GENERIC_PTR channel, ADC_TRIGGER_MASK mask) {
    _mqx_int i;

	_int_disable();

    if (channel) {
	    channel->runtime_flags |= ADC_CHANNEL_RESUMED;
    }
    else {
        for (i = 0; i < ADC_MAX_CHANNELS; i++) {
            if (adc_ch[i] && (adc_ch[i]->g.trigger & mask)) {
	            adc_ch[i]->g.runtime_flags |= ADC_CHANNEL_RESUMED;
            }
        }
    }

    _int_enable();
    return IO_OK;
}

/*FUNCTION*****************************************************************
* 
* Function Name    : _adc_trigger
* Returned Value   : IO_OK
* Comments         : Stops (resets) measurement of single channel, or more
*                    channels defined by trigger mask
*
*END*********************************************************************/

_mqx_int _adc_stop(ADC_CHANNEL_GENERIC_PTR channel, ADC_TRIGGER_MASK mask) {
    _mqx_int i;
    _mqx_uint time;

	_int_disable();
	time = _adc_hw_get_time();
	    
    if (channel) {
	    channel->runtime_flags &= ~ADC_CHANNEL_RUNNING;
    }
    else {
        for (i = 0; i < ADC_MAX_CHANNELS; i++) {
            if (adc_ch[i] && (adc_ch[i]->g.trigger & mask)) {
                    adc_ch[i]->g.runtime_flags &= ~ADC_CHANNEL_RUNNING;
            }
        }
    }
    
    running_mask = 0;
    _adc_scheduler(time);

    _int_enable();
    return IO_OK;
}


#endif // BSPCFG_ENABLE_ADC
